@@ -13,28 +13,30 @@ module Hledger.Data (
                module Hledger.Data.Amount,
                module Hledger.Data.Commodity,
                module Hledger.Data.Dates,
-               module Hledger.Data.Transaction,
-               module Hledger.Data.Ledger,
                module Hledger.Data.Journal,
+               module Hledger.Data.Ledger,
+               module Hledger.Data.Matching,
                module Hledger.Data.Posting,
                module Hledger.Data.TimeLog,
+               module Hledger.Data.Transaction,
                module Hledger.Data.Types,
-               module Hledger.Data.Utils,
                tests_Hledger_Data
               )
 where
+import Test.HUnit
+
 import Hledger.Data.Account
 import Hledger.Data.AccountName
 import Hledger.Data.Amount
 import Hledger.Data.Commodity
 import Hledger.Data.Dates
-import Hledger.Data.Transaction
-import Hledger.Data.Ledger
 import Hledger.Data.Journal
+import Hledger.Data.Ledger
+import Hledger.Data.Matching
 import Hledger.Data.Posting
 import Hledger.Data.TimeLog
+import Hledger.Data.Transaction
 import Hledger.Data.Types
-import Hledger.Data.Utils
 
 tests_Hledger_Data = TestList
     [
@@ -45,9 +47,9 @@ tests_Hledger_Data = TestList
     ,tests_Hledger_Data_Dates
     ,tests_Hledger_Data_Journal
     ,tests_Hledger_Data_Ledger
+    ,tests_Hledger_Data_Matching
     ,tests_Hledger_Data_Posting
     ,tests_Hledger_Data_TimeLog
     ,tests_Hledger_Data_Transaction
     -- ,tests_Hledger_Data_Types
-    -- ,tests_Hledger_Data_Utils
     ]
