@@ -23,13 +23,22 @@ quarterly, etc.
 module Hledger.Data.Dates
 where
 
+import Control.Monad
+import Data.List
+import Data.Maybe
 import Data.Time.Format
+import Data.Time.Calendar
 import Data.Time.Calendar.OrdinalDate
+import Data.Time.Clock
+import Data.Time.LocalTime
 import Safe (readMay)
 import System.Locale (defaultTimeLocale)
+import Test.HUnit
 import Text.ParserCombinators.Parsec
+import Text.Printf
+
 import Hledger.Data.Types
-import Hledger.Data.Utils
+import Hledger.Utils
 
 
 showDate :: Day -> String
@@ -41,6 +50,12 @@ getCurrentDay = do
     t <- getZonedTime
     return $ localDay (zonedTimeToLocalTime t)
 
+-- | Get the current local month number.
+getCurrentMonth :: IO Int
+getCurrentMonth = do
+  (_,m,_) <- toGregorian `fmap` getCurrentDay
+  return m
+
 -- | Get the current local year.
 getCurrentYear :: IO Integer
 getCurrentYear = do
@@ -106,6 +121,9 @@ orDatesFrom (DateSpan a1 b1) (DateSpan a2 b2) = DateSpan a b
 parsePeriodExpr :: Day -> String -> Either ParseError (Interval, DateSpan)
 parsePeriodExpr refdate = parsewith (periodexpr refdate)
 
+maybePeriod :: Day -> String -> Maybe (Interval,DateSpan)
+maybePeriod refdate = either (const Nothing) Just . parsePeriodExpr refdate
+
 -- | Show a DateSpan as a human-readable pseudo-period-expression string.
 dateSpanAsText :: DateSpan -> String
 dateSpanAsText (DateSpan Nothing Nothing)   = "all"
@@ -157,13 +175,19 @@ showDay day = printf "%04d/%02d/%02d" y m d where (y,m,d) = toGregorian day
 -- | Convert a smart date string to an explicit yyyy\/mm\/dd string using
 -- the provided reference date, or raise an error.
 fixSmartDateStr :: Day -> String -> String
-fixSmartDateStr t s = either parseerror id $ fixSmartDateStrEither t s
+fixSmartDateStr d s = either
+                       (\e->error' $ printf "could not parse date %s %s" (show s) (show e))
+                       id
+                       $ fixSmartDateStrEither d s
 
 -- | A safe version of fixSmartDateStr.
 fixSmartDateStrEither :: Day -> String -> Either ParseError String
-fixSmartDateStrEither t s = case parsewith smartdateonly (lowercase s) of
-                              Right sd -> Right $ showDay $ fixSmartDate t sd
-                              Left e -> Left e
+fixSmartDateStrEither d = either Left (Right . showDate) . fixSmartDateStrEither' d
+
+fixSmartDateStrEither' :: Day -> String -> Either ParseError Day
+fixSmartDateStrEither' d s = case parsewith smartdateonly (lowercase s) of
+                               Right sd -> Right $ fixSmartDate d sd
+                               Left e -> Left e
 
 -- | Convert a SmartDate to an absolute date using the provided reference date.
 fixSmartDate :: Day -> SmartDate -> Day
@@ -537,12 +561,16 @@ justdatespan rdate = do
   d <- smartdate
   return $ spanFromSmartDate rdate d
 
+-- | Make a datespan from two valid date strings parseable by parsedate
+-- (or raise an error). Eg: mkdatespan \"2011/1/1\" \"2011/12/31\".
 mkdatespan :: String -> String -> DateSpan
 mkdatespan b = DateSpan (Just $ parsedate b) . Just . parsedate
 
+nulldatespan :: DateSpan
 nulldatespan = DateSpan Nothing Nothing
 
-nulldate = parsedate "1900/01/01"
+nulldate :: Day
+nulldate = parsedate "0000/00/00"
 
 tests_Hledger_Data_Dates = TestList
  [
