@@ -8,9 +8,13 @@ plus a date and optional metadata like description and cleared status.
 
 module Hledger.Data.Transaction
 where
+import Data.List
+import Data.Maybe
+import Test.HUnit
+import Text.Printf
 import qualified Data.Map as Map
 
-import Hledger.Data.Utils
+import Hledger.Utils
 import Hledger.Data.Types
 import Hledger.Data.Dates
 import Hledger.Data.Posting
@@ -91,7 +95,7 @@ showTransaction' elide effective t =
                 where w = maximum $ map (length . paccount) ps
                       showstatus p = if pstatus p then "* " else ""
             showamt =
-                padleft 12 . showMixedAmountOrZero
+                padleft 12 . showMixedAmount
             showcomment s = if null s then "" else "  ; "++s
 
 -- | Show an account name, clipped to the given width if any, and
@@ -106,6 +110,9 @@ showAccountName w = fmt
       parenthesise s = "("++s++")"
       bracket s = "["++s++"]"
 
+hasRealPostings :: Transaction -> Bool
+hasRealPostings = not . null . realPostings
+
 realPostings :: Transaction -> [Posting]
 realPostings = filter isReal . tpostings
 
@@ -115,6 +122,9 @@ virtualPostings = filter isVirtual . tpostings
 balancedVirtualPostings :: Transaction -> [Posting]
 balancedVirtualPostings = filter isBalancedVirtual . tpostings
 
+transactionsPostings :: [Transaction] -> [Posting]
+transactionsPostings = concat . map tpostings
+
 -- | Get the sums of a transaction's real, virtual, and balanced virtual postings.
 transactionPostingBalances :: Transaction -> (MixedAmount,MixedAmount,MixedAmount)
 transactionPostingBalances t = (sumPostings $ realPostings t
@@ -130,8 +140,8 @@ isTransactionBalanced canonicalcommoditymap t =
     isZeroMixedAmount rsum' && isZeroMixedAmount bvsum'
     where
       (rsum, _, bvsum) = transactionPostingBalances t
-      rsum'  = canonicaliseMixedAmount canonicalcommoditymap $ costOfMixedAmount rsum
-      bvsum' = canonicaliseMixedAmount canonicalcommoditymap $ costOfMixedAmount bvsum
+      rsum'  = canonicaliseMixedAmountCommodity canonicalcommoditymap $ costOfMixedAmount rsum
+      bvsum' = canonicaliseMixedAmountCommodity canonicalcommoditymap $ costOfMixedAmount bvsum
 
 -- | Ensure this transaction is balanced, possibly inferring a missing
 -- amount or a conversion price first, or return an error message.
@@ -179,9 +189,10 @@ balanceTransaction canonicalcommoditymap t@Transaction{tpostings=ps}
                 where
                   conversionprice c | c == unpricedcommodity
                                         -- assign a balancing price. Use @@ for more exact output when possible.
+                                        -- invariant: prices should always be positive. Enforced with "abs"
                                         = if length ramountsinunpricedcommodity == 1
-                                           then Just $ TotalPrice $ Mixed [setAmountPrecision maxprecision $ negate $ targetcommodityamount]
-                                           else Just $ UnitPrice $ Mixed [setAmountPrecision maxprecision $ negate $ targetcommodityamount `divideAmount` (quantity unpricedamount)]
+                                           then Just $ TotalPrice $ Mixed [setAmountPrecision maxprecision $ abs $ targetcommodityamount]
+                                           else Just $ UnitPrice $ Mixed [setAmountPrecision maxprecision $ abs $ targetcommodityamount `divideAmount` (quantity unpricedamount)]
                                     | otherwise = Nothing
                       where
                         unpricedcommodity     = head $ filter (`elem` (map commodity rsumamounts)) rcommoditiesinorder
@@ -204,8 +215,8 @@ balanceTransaction canonicalcommoditymap t@Transaction{tpostings=ps}
                 where
                   conversionprice c | c == unpricedcommodity
                                         = if length bvamountsinunpricedcommodity == 1
-                                           then Just $ TotalPrice $ Mixed [setAmountPrecision maxprecision $ negate $ targetcommodityamount]
-                                           else Just $ UnitPrice $ Mixed [setAmountPrecision maxprecision $ negate $ targetcommodityamount `divideAmount` (quantity unpricedamount)]
+                                           then Just $ TotalPrice $ Mixed [setAmountPrecision maxprecision $ abs $ targetcommodityamount]
+                                           else Just $ UnitPrice $ Mixed [setAmountPrecision maxprecision $ abs $ targetcommodityamount `divideAmount` (quantity unpricedamount)]
                                     | otherwise = Nothing
                       where
                         unpricedcommodity     = head $ filter (`elem` (map commodity bvsumamounts)) bvcommoditiesinorder
