@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP #-}
 {-|
 
 A reader for hledger's (and c++ ledger's) journal file format.
@@ -110,22 +109,34 @@ module Hledger.Read.JournalReader (
        ledgeraccountname,
        ledgerdatetime,
        ledgerDefaultYear,
-       ledgerExclamationDirective,
+       ledgerDirective,
        ledgerHistoricalPrice,
        reader,
        someamount,
        tests_Hledger_Read_JournalReader
 )
 where
-import Control.Monad.Error (ErrorT(..), throwError, catchError)
+import Control.Monad
+import Control.Monad.Error
+import Data.Char (isNumber)
+import Data.List
 import Data.List.Split (wordsBy)
+import Data.Maybe
+import Data.Time.Calendar
+-- import Data.Time.Clock
+-- import Data.Time.Format
+import Data.Time.LocalTime
 import Safe (headDef)
+-- import System.Locale (defaultTimeLocale)
+import Test.HUnit
 import Text.ParserCombinators.Parsec hiding (parse)
+import Text.Printf
 
 import Hledger.Data
-import Prelude hiding (readFile)
-import Hledger.Data.UTF8 (readFile)
 import Hledger.Read.Utils
+import Hledger.Utils
+import Prelude hiding (readFile)
+import Hledger.Utils.UTF8 (readFile)
 
 
 -- let's get to it
@@ -158,22 +169,13 @@ journalFile = do
       -- As all journal line types can be distinguished by the first
       -- character, excepting transactions versus empty (blank or
       -- comment-only) lines, can use choice w/o try
-      journalItem = choice [ ledgerExclamationDirective
-                          , liftM (return . addTransaction) ledgerTransaction
-                          , liftM (return . addModifierTransaction) ledgerModifierTransaction
-                          , liftM (return . addPeriodicTransaction) ledgerPeriodicTransaction
-                          , liftM (return . addHistoricalPrice) ledgerHistoricalPrice
-                          , ledgerDefaultYear
-                          , ledgerDefaultCommodity
-                          , ledgerCommodityConversion
-                          , ledgerIgnoredPriceCommodity
-                          , ledgerTagDirective
-                          , ledgerEndTagDirective
-                          , emptyLine >> return (return id)
-                          ] <?> "journal transaction or directive"
-
-journalAddFile :: (FilePath,String) -> Journal -> Journal
-journalAddFile f j@Journal{files=fs} = j{files=fs++[f]}
+      journalItem = choice [ ledgerDirective
+                           , liftM (return . addTransaction) ledgerTransaction
+                           , liftM (return . addModifierTransaction) ledgerModifierTransaction
+                           , liftM (return . addPeriodicTransaction) ledgerPeriodicTransaction
+                           , liftM (return . addHistoricalPrice) ledgerHistoricalPrice
+                           , emptyLine >> return (return id)
+                           ] <?> "journal transaction or directive"
 
 emptyLine :: GenParser Char JournalContext ()
 emptyLine = do many spacenonewline
@@ -197,18 +199,27 @@ ledgercommentline = do
   return s
   <?> "comment"
 
-ledgerExclamationDirective :: GenParser Char JournalContext JournalUpdate
-ledgerExclamationDirective = do
-  char '!' <?> "directive"
-  directive <- many nonspace
-  case directive of
-    "include" -> ledgerInclude
-    "account" -> ledgerAccountBegin
-    "end"     -> ledgerAccountEnd
-    _         -> mzero
+ledgerDirective :: GenParser Char JournalContext JournalUpdate
+ledgerDirective = do
+  optional $ char '!'
+  choice' [
+    ledgerInclude
+   ,ledgerAlias
+   ,ledgerEndAliases
+   ,ledgerAccountBegin
+   ,ledgerAccountEnd
+   ,ledgerTagDirective
+   ,ledgerEndTagDirective
+   ,ledgerDefaultYear
+   ,ledgerDefaultCommodity
+   ,ledgerCommodityConversion
+   ,ledgerIgnoredPriceCommodity
+   ]
+  <?> "directive"
 
 ledgerInclude :: GenParser Char JournalContext JournalUpdate
 ledgerInclude = do
+  string "include"
   many1 spacenonewline
   filename <- restofline
   outerState <- getState
@@ -223,31 +234,75 @@ ledgerInclude = do
                 ErrorT $ liftM Right (readFile fp) `catch`
                   \err -> return $ Left $ printf "%s reading %s:\n%s" (show pos) fp (show err)
 
+journalAddFile :: (FilePath,String) -> Journal -> Journal
+journalAddFile f j@Journal{files=fs} = j{files=fs++[f]}
+
 ledgerAccountBegin :: GenParser Char JournalContext JournalUpdate
-ledgerAccountBegin = do many1 spacenonewline
-                        parent <- ledgeraccountname
-                        newline
-                        pushParentAccount parent
-                        return $ return id
+ledgerAccountBegin = do
+  string "account"
+  many1 spacenonewline
+  parent <- ledgeraccountname
+  newline
+  pushParentAccount parent
+  return $ return id
 
 ledgerAccountEnd :: GenParser Char JournalContext JournalUpdate
-ledgerAccountEnd = popParentAccount >> return (return id)
+ledgerAccountEnd = do
+  string "end"
+  popParentAccount
+  return (return id)
 
-ledgerModifierTransaction :: GenParser Char JournalContext ModifierTransaction
-ledgerModifierTransaction = do
-  char '=' <?> "modifier transaction"
-  many spacenonewline
-  valueexpr <- restofline
-  postings <- ledgerpostings
-  return $ ModifierTransaction valueexpr postings
+ledgerAlias :: GenParser Char JournalContext JournalUpdate
+ledgerAlias = do
+  string "alias"
+  many1 spacenonewline
+  orig <- many1 $ noneOf "="
+  char '='
+  alias <- restofline
+  addAccountAlias (accountNameWithoutPostingType $ strip orig
+                  ,accountNameWithoutPostingType $ strip alias)
+  return $ return id
 
-ledgerPeriodicTransaction :: GenParser Char JournalContext PeriodicTransaction
-ledgerPeriodicTransaction = do
-  char '~' <?> "periodic transaction"
+ledgerEndAliases :: GenParser Char JournalContext JournalUpdate
+ledgerEndAliases = do
+  string "end aliases"
+  clearAccountAliases
+  return (return id)
+
+ledgerTagDirective :: GenParser Char JournalContext JournalUpdate
+ledgerTagDirective = do
+  string "tag" <?> "tag directive"
+  many1 spacenonewline
+  _ <- many1 nonspace
+  restofline
+  return $ return id
+
+ledgerEndTagDirective :: GenParser Char JournalContext JournalUpdate
+ledgerEndTagDirective = do
+  (string "end tag" <|> string "pop") <?> "end tag or pop directive"
+  restofline
+  return $ return id
+
+ledgerDefaultYear :: GenParser Char JournalContext JournalUpdate
+ledgerDefaultYear = do
+  char 'Y' <?> "default year"
   many spacenonewline
-  periodexpr <- restofline
-  postings <- ledgerpostings
-  return $ PeriodicTransaction periodexpr postings
+  y <- many1 digit
+  let y' = read y
+  failIfInvalidYear y
+  setYear y'
+  return $ return id
+
+ledgerDefaultCommodity :: GenParser Char JournalContext JournalUpdate
+ledgerDefaultCommodity = do
+  char 'D' <?> "default commodity"
+  many1 spacenonewline
+  a <- someamount
+  -- someamount always returns a MixedAmount containing one Amount, but let's be safe
+  let as = amounts a
+  when (not $ null as) $ setCommodity $ commodity $ head as
+  restofline
+  return $ return id
 
 ledgerHistoricalPrice :: GenParser Char JournalContext HistoricalPrice
 ledgerHistoricalPrice = do
@@ -281,41 +336,21 @@ ledgerCommodityConversion = do
   restofline
   return $ return id
 
-ledgerTagDirective :: GenParser Char JournalContext JournalUpdate
-ledgerTagDirective = do
-  string "tag" <?> "tag directive"
-  many1 spacenonewline
-  _ <- many1 nonspace
-  restofline
-  return $ return id
-
-ledgerEndTagDirective :: GenParser Char JournalContext JournalUpdate
-ledgerEndTagDirective = do
-  string "end tag" <?> "end tag directive"
-  restofline
-  return $ return id
-
--- like ledgerAccountBegin, updates the JournalContext
-ledgerDefaultYear :: GenParser Char JournalContext JournalUpdate
-ledgerDefaultYear = do
-  char 'Y' <?> "default year"
+ledgerModifierTransaction :: GenParser Char JournalContext ModifierTransaction
+ledgerModifierTransaction = do
+  char '=' <?> "modifier transaction"
   many spacenonewline
-  y <- many1 digit
-  let y' = read y
-  failIfInvalidYear y
-  setYear y'
-  return $ return id
+  valueexpr <- restofline
+  postings <- ledgerpostings
+  return $ ModifierTransaction valueexpr postings
 
-ledgerDefaultCommodity :: GenParser Char JournalContext JournalUpdate
-ledgerDefaultCommodity = do
-  char 'D' <?> "default commodity"
-  many1 spacenonewline
-  a <- someamount
-  -- someamount always returns a MixedAmount containing one Amount, but let's be safe
-  let as = amounts a
-  when (not $ null as) $ setCommodity $ commodity $ head as
-  restofline
-  return $ return id
+ledgerPeriodicTransaction :: GenParser Char JournalContext PeriodicTransaction
+ledgerPeriodicTransaction = do
+  char '~' <?> "periodic transaction"
+  many spacenonewline
+  periodexpr <- restofline
+  postings <- ledgerpostings
+  return $ PeriodicTransaction periodexpr postings
 
 -- | Parse a (possibly unbalanced) ledger transaction.
 ledgerTransaction :: GenParser Char JournalContext Transaction
@@ -332,6 +367,8 @@ ledgerTransaction = do
   postings <- ledgerpostings
   return $ txnTieKnot $ Transaction date edate status code description comment md postings ""
 
+-- | Parse a date in YYYY/MM/DD format. Fewer digits are allowed. The year
+-- may be omitted if a default year has already been set.
 ledgerdate :: GenParser Char JournalContext Day
 ledgerdate = do
   -- hacky: try to ensure precise errors for invalid dates
@@ -340,28 +377,48 @@ ledgerdate = do
   datestr <- many1 $ choice' [digit, datesepchar]
   let dateparts = wordsBy (`elem` datesepchars) datestr
   currentyear <- getYear
-  let [y,m,d] = case (dateparts,currentyear) of
-                  ([m,d],Just y)  -> [show y,m,d]
-                  ([_,_],Nothing) -> fail $ "partial date "++datestr++" found, but the current year is unknown"
-                  _               -> dateparts
-      maybedate = fromGregorianValid (read y) (read m) (read d)
+  [y,m,d] <- case (dateparts,currentyear) of
+              ([m,d],Just y)  -> return [show y,m,d]
+              ([_,_],Nothing) -> fail $ "partial date "++datestr++" found, but the current year is unknown"
+              ([y,m,d],_)     -> return [y,m,d]
+              _               -> fail $ "bad date: " ++ datestr
+  let maybedate = fromGregorianValid (read y) (read m) (read d)
   case maybedate of
     Nothing   -> fail $ "bad date: " ++ datestr
     Just date -> return date
   <?> "full or partial date"
 
+-- | Parse a date and time in YYYY/MM/DD HH:MM[:SS][+-ZZZZ] format.  Any
+-- timezone will be ignored; the time is treated as local time.  Fewer
+-- digits are allowed, except in the timezone. The year may be omitted if
+-- a default year has already been set.
 ledgerdatetime :: GenParser Char JournalContext LocalTime
 ledgerdatetime = do 
   day <- ledgerdate
   many1 spacenonewline
   h <- many1 digit
+  let h' = read h
+  guard $ h' >= 0 && h' <= 23
   char ':'
   m <- many1 digit
-  s <- optionMaybe $ do
-      char ':'
-      many1 digit
-  let tod = TimeOfDay (read h) (read m) (maybe 0 (fromIntegral.read) s)
-  return $ LocalTime day tod
+  let m' = read m
+  guard $ m' >= 0 && m' <= 59
+  s <- optionMaybe $ char ':' >> many1 digit
+  let s' = case s of Just sstr -> read sstr
+                     Nothing   -> 0
+  guard $ s' >= 0 && s' <= 59
+  {- tz <- -}
+  optionMaybe $ do
+                   plusminus <- oneOf "-+"
+                   d1 <- digit
+                   d2 <- digit
+                   d3 <- digit
+                   d4 <- digit
+                   return $ plusminus:d1:d2:d3:d4:""
+  -- ltz <- liftIO $ getCurrentTimeZone
+  -- let tz' = maybe ltz (fromMaybe ltz . parseTime defaultTimeLocale "%z") tz
+  -- return $ localTimeToUTC tz' $ LocalTime day $ TimeOfDay h' m' (fromIntegral s')
+  return $ LocalTime day $ TimeOfDay h' m' (fromIntegral s')
 
 ledgereffectivedate :: Day -> GenParser Char JournalContext Day
 ledgereffectivedate actualdate = do
@@ -433,8 +490,8 @@ ledgerposting = do
   many1 spacenonewline
   status <- ledgerstatus
   many spacenonewline
-  account <- transactionaccountname
-  let (ptype, account') = (postingTypeFromAccountName account, unbracket account)
+  account <- modifiedaccountname
+  let (ptype, account') = (accountNamePostingType account, unbracket account)
   amount <- postingamount
   many spacenonewline
   comment <- ledgercomment <|> return ""
@@ -442,9 +499,14 @@ ledgerposting = do
   md <- ledgermetadata
   return (Posting status account' amount comment ptype md Nothing)
 
--- qualify with the parent account from parsing context
-transactionaccountname :: GenParser Char JournalContext AccountName
-transactionaccountname = liftM2 (++) getParentAccount ledgeraccountname
+-- Parse an account name, then apply any parent account prefix and/or account aliases currently in effect.
+modifiedaccountname :: GenParser Char JournalContext AccountName
+modifiedaccountname = do
+  a <- ledgeraccountname
+  prefix <- getParentAccount
+  let prefixed = prefix `joinAccountNames` a
+  aliases <- getAccountAliases
+  return $ accountNameApplyAliases aliases prefixed
 
 -- | Parse an account name. Account names may have single spaces inside
 -- them, and are terminated by two or more spaces. They should have one or
@@ -668,16 +730,37 @@ tests_Hledger_Read_JournalReader = TestList [
   ,"ledgerPeriodicTransaction" ~: do
      assertParse (parseWithCtx nullctx ledgerPeriodicTransaction "~ (some period expr)\n some:postings  1\n")
 
-  ,"ledgerExclamationDirective" ~: do
-     assertParse (parseWithCtx nullctx ledgerExclamationDirective "!include /some/file.x\n")
-     assertParse (parseWithCtx nullctx ledgerExclamationDirective "!account some:account\n")
-     assertParse (parseWithCtx nullctx (ledgerExclamationDirective >> ledgerExclamationDirective) "!account a\n!end\n")
+  ,"ledgerDirective" ~: do
+     assertParse (parseWithCtx nullctx ledgerDirective "!include /some/file.x\n")
+     assertParse (parseWithCtx nullctx ledgerDirective "account some:account\n")
+     assertParse (parseWithCtx nullctx (ledgerDirective >> ledgerDirective) "!account a\nend\n")
 
   ,"ledgercommentline" ~: do
      assertParse (parseWithCtx nullctx ledgercommentline "; some comment \n")
      assertParse (parseWithCtx nullctx ledgercommentline " \t; x\n")
      assertParse (parseWithCtx nullctx ledgercommentline ";x")
 
+  ,"ledgerdate" ~: do
+     assertParse (parseWithCtx nullctx ledgerdate "2011/1/1")
+     assertParseFailure (parseWithCtx nullctx ledgerdate "1/1")
+     assertParse (parseWithCtx nullctx{ctxYear=Just 2011} ledgerdate "1/1")
+
+  ,"ledgerdatetime" ~: do
+      let p = do {t <- ledgerdatetime; eof; return t}
+          bad = assertParseFailure . parseWithCtx nullctx p
+          good = assertParse . parseWithCtx nullctx p
+      bad "2011/1/1"
+      bad "2011/1/1 24:00:00"
+      bad "2011/1/1 00:60:00"
+      bad "2011/1/1 00:00:60"
+      good "2011/1/1 00:00"
+      good "2011/1/1 23:59:59"
+      good "2011/1/1 3:5:7"
+      -- timezone is parsed but ignored
+      let startofday = LocalTime (fromGregorian 2011 1 1) (TimeOfDay 0 0 (fromIntegral 0))
+      assertParseEqual (parseWithCtx nullctx p "2011/1/1 00:00-0800") startofday
+      assertParseEqual (parseWithCtx nullctx p "2011/1/1 00:00+1234") startofday
+
   ,"ledgerDefaultYear" ~: do
      assertParse (parseWithCtx nullctx ledgerDefaultYear "Y 2010\n")
      assertParse (parseWithCtx nullctx ledgerDefaultYear "Y 10001\n")
@@ -699,6 +782,8 @@ tests_Hledger_Read_JournalReader = TestList [
 
   ,"ledgerEndTagDirective" ~: do
      assertParse (parseWithCtx nullctx ledgerEndTagDirective "end tag \n")
+  ,"ledgerEndTagDirective" ~: do
+     assertParse (parseWithCtx nullctx ledgerEndTagDirective "pop \n")
 
   ,"ledgeraccountname" ~: do
     assertBool "ledgeraccountname parses a normal accountname" (isRight $ parsewith ledgeraccountname "a:b:c")
